@@ -19,7 +19,6 @@ import re
 import string
 import tempfile
 
-from swift import gettext_ as _
 from swift.common.middleware.x_profile.exceptions import PLOTLIBNotInstalled
 from swift.common.middleware.x_profile.exceptions import ODFLIBNotInstalled
 from swift.common.middleware.x_profile.exceptions import NotFoundException
@@ -246,7 +245,7 @@ class HTMLViewer(object):
         if multiple:
             return value
         if isinstance(value, list):
-            return int(value[0]) if isinstance(default, int) else value[0]
+            return eval(value[0]) if isinstance(default, int) else value[0]
         else:
             return value
 
@@ -307,7 +306,7 @@ class HTMLViewer(object):
                                                  nfl_filter, download_format)
             headers.append(('Access-Control-Allow-Origin', '*'))
         else:
-            raise MethodNotAllowed(_('method %s is not allowed.') % method)
+            raise MethodNotAllowed('method %s is not allowed.' % method)
         return content, headers
 
     def index_page(self, log_files=None, sort='time', limit=-1,
@@ -318,7 +317,7 @@ class HTMLViewer(object):
         try:
             stats = Stats2(*log_files)
         except (IOError, ValueError):
-            raise DataLoadFailure(_('Can not load profile data from %s.')
+            raise DataLoadFailure('Can not load profile data from %s.'
                                   % log_files)
         if not fulldirs:
             stats.strip_dirs()
@@ -369,7 +368,7 @@ class HTMLViewer(object):
     def download(self, log_files, sort='time', limit=-1, nfl_filter='',
                  output_format='default'):
         if len(log_files) == 0:
-            raise NotFoundException(_('no log file found'))
+            raise NotFoundException('no log file found')
         try:
             nfl_esc = nfl_filter.replace(r'(', r'\(').replace(r')', r'\)')
             # remove the slash that is intentionally added in the URL
@@ -392,14 +391,14 @@ class HTMLViewer(object):
         except ODFLIBNotInstalled:
             raise
         except Exception as ex:
-            raise ProfileException(_('Data download error: %s') % ex)
+            raise ProfileException('Data download error: %s' % ex)
 
     def plot(self, log_files, sort='time', limit=10, nfl_filter='',
              metric_selected='cc', plot_type='bar'):
         if not PLOTLIB_INSTALLED:
-            raise PLOTLIBNotInstalled(_('python-matplotlib not installed.'))
+            raise PLOTLIBNotInstalled('python-matplotlib not installed.')
         if len(log_files) == 0:
-            raise NotFoundException(_('no log file found'))
+            raise NotFoundException('no log file found')
         try:
             stats = Stats2(*log_files)
             stats.sort_stats(sort)
@@ -433,7 +432,7 @@ class HTMLViewer(object):
                 data = profile_img.read()
                 return data, [('content-type', 'image/jpg')]
         except Exception as ex:
-            raise ProfileException(_('plotting results failed due to %s') % ex)
+            raise ProfileException('plotting results failed due to %s' % ex)
 
     def format_source_code(self, nfl):
         nfls = re.split('[:()]', nfl)
@@ -444,7 +443,7 @@ class HTMLViewer(object):
             lineno = 0
         # for security reason, this need to be fixed.
         if not file_path.endswith('.py'):
-            return _('The file type are forbidden to access!')
+            return 'The file type are forbidden to access!'
         try:
             data = []
             i = 0
@@ -465,7 +464,7 @@ class HTMLViewer(object):
                         data.append(fmt % (i, i, i, el))
             data = ''.join(data)
         except Exception:
-            return _('Can not access the file %s.') % file_path
+            return 'Can not access the file %s.' % file_path
         return '<pre>%s</pre>' % data
 
     def generate_stats_html(self, stats, app_path, profile_id, *selection):
