@@ -15,17 +15,18 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
-import unittest2
+import unittest
 import json
 from uuid import uuid4
-from unittest2 import SkipTest
 from string import ascii_letters
 
+import six
 from six.moves import range
 from swift.common.middleware.acl import format_acl
+from swift.common.utils import distribute_evenly
 
 from test.functional import check_response, retry, requires_acls, \
-    load_constraint
+    load_constraint, SkipTest
 import test.functional as tf
 
 
@@ -37,7 +38,55 @@ def tearDownModule():
     tf.teardown_package()
 
 
-class TestAccount(unittest2.TestCase):
+class TestAccount(unittest.TestCase):
+    existing_metadata = None
+
+    @classmethod
+    def get_meta(cls):
+        def head(url, token, parsed, conn):
+            conn.request('HEAD', parsed.path, '', {'X-Auth-Token': token})
+            return check_response(conn)
+        resp = retry(head)
+        resp.read()
+        return dict((k, v) for k, v in resp.getheaders() if
+                    k.lower().startswith('x-account-meta'))
+
+    @classmethod
+    def clear_meta(cls, remove_metadata_keys):
+        def post(url, token, parsed, conn, hdr_keys):
+            headers = {'X-Auth-Token': token}
+            headers.update((k, '') for k in hdr_keys)
+            conn.request('POST', parsed.path, '', headers)
+            return check_response(conn)
+
+        buckets = (len(remove_metadata_keys) - 1) // 90 + 1
+        for batch in distribute_evenly(remove_metadata_keys, buckets):
+            resp = retry(post, batch)
+            resp.read()
+
+    @classmethod
+    def set_meta(cls, metadata):
+        def post(url, token, parsed, conn, meta_hdrs):
+            headers = {'X-Auth-Token': token}
+            headers.update(meta_hdrs)
+            conn.request('POST', parsed.path, '', headers)
+            return check_response(conn)
+
+        if not metadata:
+            return
+        resp = retry(post, metadata)
+        resp.read()
+
+    @classmethod
+    def setUpClass(cls):
+        # remove and stash any existing account user metadata before tests
+        cls.existing_metadata = cls.get_meta()
+        cls.clear_meta(cls.existing_metadata.keys())
+
+    @classmethod
+    def tearDownClass(cls):
+        # replace any stashed account user metadata
+        cls.set_meta(cls.existing_metadata)
 
     def setUp(self):
         self.max_meta_count = load_constraint('max_meta_count')
@@ -45,35 +94,10 @@ class TestAccount(unittest2.TestCase):
         self.max_meta_overall_size = load_constraint('max_meta_overall_size')
         self.max_meta_value_length = load_constraint('max_meta_value_length')
 
-        def head(url, token, parsed, conn):
-            conn.request('HEAD', parsed.path, '', {'X-Auth-Token': token})
-            return check_response(conn)
-        resp = retry(head)
-        self.existing_metadata = set([
-            k for k, v in resp.getheaders() if
-            k.lower().startswith('x-account-meta')])
-
     def tearDown(self):
-        def head(url, token, parsed, conn):
-            conn.request('HEAD', parsed.path, '', {'X-Auth-Token': token})
-            return check_response(conn)
-        resp = retry(head)
-        resp.read()
-        new_metadata = set(
-            [k for k, v in resp.getheaders() if
-             k.lower().startswith('x-account-meta')])
-
-        def clear_meta(url, token, parsed, conn, remove_metadata_keys):
-            headers = {'X-Auth-Token': token}
-            headers.update((k, '') for k in remove_metadata_keys)
-            conn.request('POST', parsed.path, '', headers)
-            return check_response(conn)
-        extra_metadata = list(self.existing_metadata ^ new_metadata)
-        for i in range(0, len(extra_metadata), 90):
-            batch = extra_metadata[i:i + 90]
-            resp = retry(clear_meta, batch)
-            resp.read()
-            self.assertEqual(resp.status // 100, 2)
+        # clean up any account user metadata created by test
+        new_metadata = self.get_meta().keys()
+        self.clear_meta(new_metadata)
 
     def test_metadata(self):
         if tf.skip:
@@ -98,11 +122,11 @@ class TestAccount(unittest2.TestCase):
         resp = retry(head)
         resp.read()
         self.assertIn(resp.status, (200, 204))
-        self.assertEqual(resp.getheader('x-account-meta-test'), None)
+        self.assertIsNone(resp.getheader('x-account-meta-test'))
         resp = retry(get)
         resp.read()
         self.assertIn(resp.status, (200, 204))
-        self.assertEqual(resp.getheader('x-account-meta-test'), None)
+        self.assertIsNone(resp.getheader('x-account-meta-test'))
         resp = retry(post, 'Value')
         resp.read()
         self.assertEqual(resp.status, 204)
@@ -126,7 +150,7 @@ class TestAccount(unittest2.TestCase):
 
         # needs to be an acceptable header size
         num_keys = 8
-        max_key_size = load_constraint('max_header_size') / num_keys
+        max_key_size = load_constraint('max_header_size') // num_keys
         acl = {'admin': [c * max_key_size for c in ascii_letters[:num_keys]]}
         headers = {'x-account-access-control': format_acl(
             version=2, acl_dict=acl)}
@@ -164,7 +188,7 @@ class TestAccount(unittest2.TestCase):
         resp = retry(post, headers, use_account=1)
         resp.read()
         self.assertEqual(resp.status, 400)
-        self.assertEqual(resp.getheader('X-Account-Access-Control'), None)
+        self.assertIsNone(resp.getheader('X-Account-Access-Control'))
 
     @requires_acls
     def test_invalid_acl_values(self):
@@ -180,7 +204,7 @@ class TestAccount(unittest2.TestCase):
         resp = retry(post, headers=headers, use_account=1)
         resp.read()
         self.assertEqual(resp.status, 400)
-        self.assertEqual(resp.getheader('X-Account-Access-Control'), None)
+        self.assertIsNone(resp.getheader('X-Account-Access-Control'))
 
     @requires_acls
     def test_read_only_acl(self):
@@ -215,7 +239,7 @@ class TestAccount(unittest2.TestCase):
         resp.read()
         self.assertIn(resp.status, (200, 204))
         # but not acls
-        self.assertEqual(resp.getheader('X-Account-Access-Control'), None)
+        self.assertIsNone(resp.getheader('X-Account-Access-Control'))
 
         # read-only can not write metadata
         headers = {'x-account-meta-test': 'value'}
@@ -266,7 +290,7 @@ class TestAccount(unittest2.TestCase):
         resp.read()
         self.assertIn(resp.status, (200, 204))
         # but not acls
-        self.assertEqual(resp.getheader('X-Account-Access-Control'), None)
+        self.assertIsNone(resp.getheader('X-Account-Access-Control'))
 
         # read-write can not write account metadata
         headers = {'x-account-meta-test': 'value'}
@@ -368,12 +392,11 @@ class TestAccount(unittest2.TestCase):
         # read-only tester3 can read account metadata
         resp = retry(get, use_account=3)
         resp.read()
-        self.assertTrue(
-            resp.status in (200, 204),
-            'Expected status in (200, 204), got %s' % resp.status)
+        self.assertIn(resp.status, (200, 204),
+                      'Expected status in (200, 204), got %s' % resp.status)
         self.assertEqual(resp.getheader('X-Account-Meta-Test'), value)
         # but not temp-url-key
-        self.assertEqual(resp.getheader('X-Account-Meta-Temp-Url-Key'), None)
+        self.assertIsNone(resp.getheader('X-Account-Meta-Temp-Url-Key'))
 
         # grant read-write access to tester3
         acl_user = tf.swift_test_user[2]
@@ -387,12 +410,11 @@ class TestAccount(unittest2.TestCase):
         # read-write tester3 can read account metadata
         resp = retry(get, use_account=3)
         resp.read()
-        self.assertTrue(
-            resp.status in (200, 204),
-            'Expected status in (200, 204), got %s' % resp.status)
+        self.assertIn(resp.status, (200, 204),
+                      'Expected status in (200, 204), got %s' % resp.status)
         self.assertEqual(resp.getheader('X-Account-Meta-Test'), value)
         # but not temp-url-key
-        self.assertEqual(resp.getheader('X-Account-Meta-Temp-Url-Key'), None)
+        self.assertIsNone(resp.getheader('X-Account-Meta-Temp-Url-Key'))
 
         # grant admin access to tester3
         acl_user = tf.swift_test_user[2]
@@ -406,9 +428,8 @@ class TestAccount(unittest2.TestCase):
         # admin tester3 can read account metadata
         resp = retry(get, use_account=3)
         resp.read()
-        self.assertTrue(
-            resp.status in (200, 204),
-            'Expected status in (200, 204), got %s' % resp.status)
+        self.assertIn(resp.status, (200, 204),
+                      'Expected status in (200, 204), got %s' % resp.status)
         self.assertEqual(resp.getheader('X-Account-Meta-Test'), value)
         # including temp-url-key
         self.assertEqual(resp.getheader('X-Account-Meta-Temp-Url-Key'),
@@ -424,9 +445,8 @@ class TestAccount(unittest2.TestCase):
         self.assertEqual(resp.status, 204)
         resp = retry(get, use_account=3)
         resp.read()
-        self.assertTrue(
-            resp.status in (200, 204),
-            'Expected status in (200, 204), got %s' % resp.status)
+        self.assertIn(resp.status, (200, 204),
+                      'Expected status in (200, 204), got %s' % resp.status)
         self.assertEqual(resp.getheader('X-Account-Meta-Temp-Url-Key'),
                          secret)
 
@@ -464,13 +484,13 @@ class TestAccount(unittest2.TestCase):
                          use_account=1)
             resp.read()
             self.assertEqual(resp.status, 204)
-            self.assertEqual(resp.getheader('X-Account-Access-Control'), None)
+            self.assertIsNone(resp.getheader('X-Account-Access-Control'))
 
             # User1 can GET their own empty account
             resp = retry(get, use_account=1)
             resp.read()
             self.assertEqual(resp.status // 100, 2)
-            self.assertEqual(resp.getheader('X-Account-Access-Control'), None)
+            self.assertIsNone(resp.getheader('X-Account-Access-Control'))
 
             # User2 can't GET User1's account
             resp = retry(get, use_account=2, url_account=1)
@@ -514,7 +534,7 @@ class TestAccount(unittest2.TestCase):
             resp = retry(head, use_account=2, url_account=1)
             resp.read()
             self.assertEqual(resp.status, 204)
-            self.assertEqual(resp.getheader('x-account-access-control'), None)
+            self.assertIsNone(resp.getheader('x-account-access-control'))
 
             # User2 can PUT and DELETE a container
             resp = retry(put, use_account=2, url_account=1,
@@ -539,7 +559,7 @@ class TestAccount(unittest2.TestCase):
             resp = retry(head, use_account=2, url_account=1)
             resp.read()
             self.assertEqual(resp.status, 204)
-            self.assertEqual(resp.getheader('x-account-access-control'), None)
+            self.assertIsNone(resp.getheader('x-account-access-control'))
 
             # User2 can't PUT a container
             resp = retry(put, use_account=2, url_account=1,
@@ -577,13 +597,13 @@ class TestAccount(unittest2.TestCase):
             resp = retry(post, headers={'X-Account-Access-Control': '{}'})
             resp.read()
             self.assertEqual(resp.status, 204)
-            self.assertEqual(resp.getheader('X-Account-Access-Control'), None)
+            self.assertIsNone(resp.getheader('X-Account-Access-Control'))
 
             # User1 can GET their own empty account
             resp = retry(get)
             resp.read()
             self.assertEqual(resp.status // 100, 2)
-            self.assertEqual(resp.getheader('X-Account-Access-Control'), None)
+            self.assertIsNone(resp.getheader('X-Account-Access-Control'))
 
             # User1 can POST non-empty data
             acl_json = '{"admin":["bob"]}'
@@ -636,13 +656,13 @@ class TestAccount(unittest2.TestCase):
             resp = retry(post, headers={'X-Account-Access-Control': '{}'})
             resp.read()
             self.assertEqual(resp.status, 204)
-            self.assertEqual(resp.getheader('X-Account-Access-Control'), None)
+            self.assertIsNone(resp.getheader('X-Account-Access-Control'))
 
             # User1 can GET their own empty account
             resp = retry(get)
             resp.read()
             self.assertEqual(resp.status // 100, 2)
-            self.assertEqual(resp.getheader('X-Account-Access-Control'), None)
+            self.assertIsNone(resp.getheader('X-Account-Access-Control'))
 
             # User1 can POST non-empty data
             acl_json = '{"admin":["bob"]}'
@@ -699,7 +719,9 @@ class TestAccount(unittest2.TestCase):
             return check_response(conn)
         uni_key = u'X-Account-Meta-uni\u0E12'
         uni_value = u'uni\u0E12'
-        if (tf.web_front_end == 'integral'):
+        # Note that py3 has issues with non-ascii header names; see
+        # https://bugs.python.org/issue37093
+        if (tf.web_front_end == 'integral' and six.PY2):
             resp = retry(post, uni_key, '1')
             resp.read()
             self.assertIn(resp.status, (201, 204))
@@ -713,9 +735,14 @@ class TestAccount(unittest2.TestCase):
         resp = retry(head)
         resp.read()
         self.assertIn(resp.status, (200, 204))
-        self.assertEqual(resp.getheader('X-Account-Meta-uni'),
-                         uni_value.encode('utf-8'))
-        if (tf.web_front_end == 'integral'):
+        if six.PY2:
+            self.assertEqual(resp.getheader('X-Account-Meta-uni'),
+                             uni_value.encode('utf8'))
+        else:
+            self.assertEqual(resp.getheader('X-Account-Meta-uni'),
+                             uni_value)
+        # See above note about py3 and non-ascii header names
+        if (tf.web_front_end == 'integral' and six.PY2):
             resp = retry(post, uni_key, uni_value)
             resp.read()
             self.assertEqual(resp.status, 204)
@@ -769,6 +796,14 @@ class TestAccount(unittest2.TestCase):
                          'k' * self.max_meta_name_length): 'v'})
         resp.read()
         self.assertEqual(resp.status, 204)
+        # Clear it, so the value-length checking doesn't accidentally trip
+        # the overall max
+        resp = retry(post,
+                     {'X-Account-Meta-' + (
+                         'k' * self.max_meta_name_length): ''})
+        resp.read()
+        self.assertEqual(resp.status, 204)
+
         resp = retry(
             post,
             {'X-Account-Meta-' + ('k' * (
@@ -798,11 +833,6 @@ class TestAccount(unittest2.TestCase):
             conn.request('POST', parsed.path, '', headers)
             return check_response(conn)
 
-        # TODO: Find the test that adds these and remove them.
-        headers = {'x-remove-account-meta-temp-url-key': 'remove',
-                   'x-remove-account-meta-temp-url-key-2': 'remove'}
-        resp = retry(post, headers)
-
         headers = {}
         for x in range(self.max_meta_count):
             headers['X-Account-Meta-%d' % x] = 'v'
@@ -820,6 +850,9 @@ class TestAccount(unittest2.TestCase):
         if tf.skip:
             raise SkipTest
 
+        if tf.in_process:
+            tf.skip_if_no_xattrs()
+
         def post(url, token, parsed, conn, extra_headers):
             headers = {'X-Auth-Token': token}
             headers.update(extra_headers)
@@ -863,7 +896,7 @@ class TestAccount(unittest2.TestCase):
         self.assertEqual(resp.status, 400)
 
 
-class TestAccountInNonDefaultDomain(unittest2.TestCase):
+class TestAccountInNonDefaultDomain(unittest.TestCase):
     def setUp(self):
         if tf.skip or tf.skip2 or tf.skip_if_not_v3:
             raise SkipTest('AUTH VERSION 3 SPECIFIC TEST')
@@ -891,5 +924,157 @@ class TestAccountInNonDefaultDomain(unittest2.TestCase):
         self.assertIn('X-Account-Project-Domain-Id', resp.headers)
 
 
+class TestAccountQuotas(unittest.TestCase):
+    def setUp(self):
+        if 'account_quotas' not in tf.cluster_info:
+            raise SkipTest('Account quotas are not enabled')
+
+        self.policies = tf.FunctionalStoragePolicyCollection.from_info()
+
+    def _check_user_cannot_post(self, headers):
+        def post(url, token, parsed, conn):
+            conn.request('POST', parsed.path, '',
+                         dict({'X-Auth-Token': token}, **headers))
+            return check_response(conn)
+
+        resp = retry(post)
+        resp.read()
+        self.assertEqual(resp.status, 403)
+
+    def test_user_cannot_set_own_quota_legacy(self):
+        self._check_user_cannot_post({'X-Account-Meta-Quota-Bytes': '0'})
+
+    def test_user_cannot_set_own_quota(self):
+        self._check_user_cannot_post({'X-Account-Quota-Bytes': '0'})
+        self._check_user_cannot_post({'X-Account-Quota-Count': '0'})
+
+    def test_user_cannot_set_own_policy_quota(self):
+        policy = self.policies.select()['name']
+        self._check_user_cannot_post(
+            {'X-Account-Quota-Bytes-Policy-' + policy: '0'})
+        self._check_user_cannot_post(
+            {'X-Account-Quota-Count-Policy-' + policy: '0'})
+
+    def test_user_cannot_remove_own_quota_legacy(self):
+        self._check_user_cannot_post(
+            {'X-Remove-Account-Meta-Quota-Bytes': 't'})
+
+    def test_user_cannot_remove_own_quota(self):
+        self._check_user_cannot_post(
+            {'X-Remove-Account-Quota-Bytes': 't'})
+        self._check_user_cannot_post(
+            {'X-Remove-Account-Quota-Count': 't'})
+
+    def test_user_cannot_remove_own_policy_quota(self):
+        policy = self.policies.select()['name']
+        self._check_user_cannot_post(
+            {'X-Remove-Account-Quota-Bytes-Policy-' + policy: 't'})
+        self._check_user_cannot_post(
+            {'X-Remove-Account-Quota-Count-Policy-' + policy: 't'})
+
+    def _check_admin_can_post(self, headers):
+        def post(url, token, parsed, conn):
+            conn.request('POST', parsed.path, '',
+                         dict({'X-Auth-Token': token}, **headers))
+            return check_response(conn)
+
+        resp = retry(post, use_account=6, url_account=1)
+        resp.read()
+        self.assertEqual(resp.status, 204)
+
+    def _test_admin_can_set_and_remove_user_quota(self, quota_header):
+        if tf.skip_if_no_reseller_admin:
+            raise SkipTest('No admin user configured')
+
+        def get_current_quota():
+            def head(url, token, parsed, conn):
+                conn.request('HEAD', parsed.path, '',
+                             {'X-Auth-Token': token})
+                return check_response(conn)
+
+            # Use user, not admin, to ensure globals in test.functional
+            # are properly populated before issuing POSTs
+            resp = retry(head)
+            resp.read()
+            self.assertEqual(resp.status, 204)
+            # Non-user-meta is authoritative now
+            return resp.headers.get(quota_header.replace('-Meta-', '-'),
+                                    resp.headers.get(quota_header))
+
+        original_quota = get_current_quota()
+
+        try:
+            self._check_admin_can_post({quota_header: '123'})
+            self.assertEqual('123', get_current_quota())
+
+            self._check_admin_can_post(
+                {quota_header.replace('X-', 'X-Remove-'): 't'})
+            self.assertIsNone(get_current_quota())
+
+            self._check_admin_can_post({quota_header: '111'})
+            self.assertEqual('111', get_current_quota())
+
+            # Can also remove with an explicit empty string
+            self._check_admin_can_post({quota_header: ''})
+            self.assertIsNone(get_current_quota())
+
+            self._check_admin_can_post({quota_header: '0'})
+            self.assertEqual('0', get_current_quota())
+        finally:
+            self._check_admin_can_post({quota_header: original_quota or ''})
+
+    def test_admin_can_set_and_remove_user_quota_legacy(self):
+        self._test_admin_can_set_and_remove_user_quota(
+            'X-Account-Meta-Quota-Bytes')
+
+    def test_admin_can_set_and_remove_user_quota(self):
+        self._test_admin_can_set_and_remove_user_quota(
+            'X-Account-Quota-Bytes')
+        self._test_admin_can_set_and_remove_user_quota(
+            'X-Account-Quota-Count')
+
+    def test_admin_can_set_and_remove_user_policy_quota(self):
+        if tf.skip_if_no_reseller_admin:
+            raise SkipTest('No admin user configured')
+        policy = self.policies.select()['name']
+
+        def get_current_quota(header):
+            def head(url, token, parsed, conn):
+                conn.request('HEAD', parsed.path, '',
+                             {'X-Auth-Token': token})
+                return check_response(conn)
+
+            # Use user, not admin, to ensure globals in test.functional
+            # are properly populated before issuing POSTs
+            resp = retry(head)
+            resp.read()
+            self.assertEqual(resp.status, 204)
+            return resp.headers.get(header)
+
+        for quota_header in ('X-Account-Quota-Bytes-Policy-' + policy,
+                             'X-Account-Quota-Count-Policy-' + policy):
+            original_quota = get_current_quota(quota_header)
+            try:
+                self._check_admin_can_post({quota_header: '123'})
+                self.assertEqual('123', get_current_quota(quota_header))
+
+                self._check_admin_can_post(
+                    {quota_header.replace('X-', 'X-Remove-'): 't'})
+                self.assertIsNone(get_current_quota(quota_header))
+
+                self._check_admin_can_post({quota_header: '111'})
+                self.assertEqual('111', get_current_quota(quota_header))
+
+                # Can also remove with an explicit empty string
+                self._check_admin_can_post({quota_header: ''})
+                self.assertIsNone(get_current_quota(quota_header))
+
+                self._check_admin_can_post({quota_header: '0'})
+                self.assertEqual('0', get_current_quota(quota_header))
+            finally:
+                self._check_admin_can_post(
+                    {quota_header: original_quota or ''})
+
+
 if __name__ == '__main__':
-    unittest2.main()
+    unittest.main()
