@@ -13,7 +13,7 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
-from datetime import datetime, timedelta
+from datetime import timedelta
 import hashlib
 from mock import patch, MagicMock
 import unittest
@@ -29,10 +29,13 @@ from swift.common.middleware.s3api.subresource import ACL, User, Owner, \
     Grant, encode_acl
 from test.unit.common.middleware.s3api.test_s3api import S3ApiTestCase
 from swift.common.middleware.s3api.s3request import S3Request, \
-    S3AclRequest, SigV4Request, SIGV4_X_AMZ_DATE_FORMAT, HashingInput
+    S3AclRequest, SigV4Request, SIGV4_X_AMZ_DATE_FORMAT, HashingInput, \
+    S3InputSHA256Mismatch
 from swift.common.middleware.s3api.s3response import InvalidArgument, \
     NoSuchBucket, InternalError, ServiceUnavailable, \
-    AccessDenied, SignatureDoesNotMatch, RequestTimeTooSkewed, BadDigest
+    AccessDenied, SignatureDoesNotMatch, RequestTimeTooSkewed, \
+    InvalidPartArgument, InvalidPartNumber, InvalidRequest, \
+    XAmzContentSHA256Mismatch
 from swift.common.utils import md5
 
 from test.debug_logger import debug_logger
@@ -95,7 +98,6 @@ class TestRequest(S3ApiTestCase):
     def setUp(self):
         super(TestRequest, self).setUp()
         self.s3api.conf.s3_acl = True
-        self.swift.s3_acl = True
 
     @patch('swift.common.middleware.s3api.acl_handlers.ACL_MAP', Fake_ACL_MAP)
     @patch('swift.common.middleware.s3api.s3request.S3AclRequest.authenticate',
@@ -117,12 +119,11 @@ class TestRequest(S3ApiTestCase):
                       'check_permission') as m_check_permission:
             mock_get_resp.return_value = fake_swift_resp \
                 or FakeResponse(self.s3api.conf.s3_acl)
-            return mock_get_resp, m_check_permission,\
+            return mock_get_resp, m_check_permission, \
                 s3_req.get_response(self.s3api)
 
     def test_get_response_without_s3_acl(self):
         self.s3api.conf.s3_acl = False
-        self.swift.s3_acl = False
         mock_get_resp, m_check_permission, s3_resp = \
             self._test_get_response('HEAD')
         self.assertFalse(hasattr(s3_resp, 'bucket_acl'))
@@ -412,7 +413,7 @@ class TestRequest(S3ApiTestCase):
                 'Signature=X' % (
                     scope_date,
                     ';'.join(sorted(['host', included_header]))),
-            'X-Amz-Content-SHA256': '0123456789'}
+            'X-Amz-Content-SHA256': '0' * 64}
 
         headers.update(date_header)
         req = Request.blank('/', environ=environ, headers=headers)
@@ -472,12 +473,12 @@ class TestRequest(S3ApiTestCase):
 
         # near-past X-Amz-Date headers
         date_header = {'X-Amz-Date': self.get_v4_amz_date_header(
-            datetime.utcnow() - timedelta(minutes=10)
+            timedelta(minutes=-10)
         )}
         self._test_request_timestamp_sigv4(date_header)
 
         date_header = {'X-Amz-Date': self.get_v4_amz_date_header(
-            datetime.utcnow() - timedelta(minutes=10)
+            timedelta(minutes=-10)
         )}
         with self.assertRaises(RequestTimeTooSkewed) as cm, \
                 patch.object(self.s3api.conf, 'allowable_clock_skew', 300):
@@ -485,19 +486,19 @@ class TestRequest(S3ApiTestCase):
 
         # near-future X-Amz-Date headers
         date_header = {'X-Amz-Date': self.get_v4_amz_date_header(
-            datetime.utcnow() + timedelta(minutes=10)
+            timedelta(minutes=10)
         )}
         self._test_request_timestamp_sigv4(date_header)
 
         date_header = {'X-Amz-Date': self.get_v4_amz_date_header(
-            datetime.utcnow() + timedelta(minutes=10)
+            timedelta(minutes=10)
         )}
         with self.assertRaises(RequestTimeTooSkewed) as cm, \
                 patch.object(self.s3api.conf, 'allowable_clock_skew', 300):
             self._test_request_timestamp_sigv4(date_header)
 
         date_header = {'X-Amz-Date': self.get_v4_amz_date_header(
-            datetime.utcnow() + timedelta(days=1)
+            timedelta(days=1)
         )}
         with self.assertRaises(RequestTimeTooSkewed) as cm:
             self._test_request_timestamp_sigv4(date_header)
@@ -594,7 +595,7 @@ class TestRequest(S3ApiTestCase):
                 'Credential=test/%s/us-east-1/s3/aws4_request, '
                 'SignedHeaders=host;x-amz-content-sha256;x-amz-date,'
                 'Signature=X' % self.get_v4_amz_date_header().split('T', 1)[0],
-            'X-Amz-Content-SHA256': '0123456789',
+            'X-Amz-Content-SHA256': '0' * 64,
             'Date': self.get_date_header(),
             'X-Amz-Date': x_amz_date}
 
@@ -604,7 +605,7 @@ class TestRequest(S3ApiTestCase):
         headers_to_sign = sigv4_req._headers_to_sign()
         self.assertEqual(headers_to_sign, [
             ('host', 'localhost:80'),
-            ('x-amz-content-sha256', '0123456789'),
+            ('x-amz-content-sha256', '0' * 64),
             ('x-amz-date', x_amz_date)])
 
         # no x-amz-date
@@ -614,7 +615,7 @@ class TestRequest(S3ApiTestCase):
                 'Credential=test/%s/us-east-1/s3/aws4_request, '
                 'SignedHeaders=host;x-amz-content-sha256,'
                 'Signature=X' % self.get_v4_amz_date_header().split('T', 1)[0],
-            'X-Amz-Content-SHA256': '0123456789',
+            'X-Amz-Content-SHA256': '1' * 64,
             'Date': self.get_date_header()}
 
         req = Request.blank('/', environ=environ, headers=headers)
@@ -623,7 +624,7 @@ class TestRequest(S3ApiTestCase):
         headers_to_sign = sigv4_req._headers_to_sign()
         self.assertEqual(headers_to_sign, [
             ('host', 'localhost:80'),
-            ('x-amz-content-sha256', '0123456789')])
+            ('x-amz-content-sha256', '1' * 64)])
 
         # SignedHeaders says, host and x-amz-date included but there is not
         # X-Amz-Date header
@@ -633,7 +634,7 @@ class TestRequest(S3ApiTestCase):
                 'Credential=test/%s/us-east-1/s3/aws4_request, '
                 'SignedHeaders=host;x-amz-content-sha256;x-amz-date,'
                 'Signature=X' % self.get_v4_amz_date_header().split('T', 1)[0],
-            'X-Amz-Content-SHA256': '0123456789',
+            'X-Amz-Content-SHA256': '2' * 64,
             'Date': self.get_date_header()}
 
         req = Request.blank('/', environ=environ, headers=headers)
@@ -730,7 +731,7 @@ class TestRequest(S3ApiTestCase):
                 'Credential=test/%s/us-east-1/s3/aws4_request, '
                 'SignedHeaders=host;x-amz-content-sha256;x-amz-date,'
                 'Signature=X' % self.get_v4_amz_date_header().split('T', 1)[0],
-            'X-Amz-Content-SHA256': '0123456789',
+            'X-Amz-Content-SHA256': '0' * 64,
             'Date': self.get_date_header(),
             'X-Amz-Date': x_amz_date}
 
@@ -872,7 +873,7 @@ class TestRequest(S3ApiTestCase):
                 'Credential=test/%s/us-east-1/s3/aws4_request, '
                 'SignedHeaders=host;x-amz-content-sha256;x-amz-date,'
                 'Signature=X' % amz_date_header.split('T', 1)[0],
-            'X-Amz-Content-SHA256': '0123456789',
+            'X-Amz-Content-SHA256': '0' * 64,
             'X-Amz-Date': amz_date_header
         })
         sigv4_req = SigV4Request(
@@ -942,18 +943,18 @@ class TestRequest(S3ApiTestCase):
         # Virtual hosted-style
         self.s3api.conf.storage_domains = ['s3.test.com']
 
-        # bad sha256
+        # bad sha256 -- but note that SHAs are not checked for GET/HEAD!
         environ = {
             'HTTP_HOST': 'bucket.s3.test.com',
-            'REQUEST_METHOD': 'GET'}
+            'REQUEST_METHOD': 'PUT'}
         headers = {
             'Authorization':
                 'AWS4-HMAC-SHA256 '
                 'Credential=test/20210104/us-east-1/s3/aws4_request, '
                 'SignedHeaders=host;x-amz-content-sha256;x-amz-date,'
-                'Signature=f721a7941d5b7710344bc62cc45f87e66f4bb1dd00d9075ee61'
-                '5b1a5c72b0f8c',
-            'X-Amz-Content-SHA256': 'bad',
+                'Signature=5f31c77dbc63e7c6ffc84dae60a9261c57c44884fe7927baeb9'
+                '84f418d4d511a',
+            'X-Amz-Content-SHA256': '0' * 64,
             'Date': 'Mon, 04 Jan 2021 10:26:23 -0000',
             'X-Amz-Date': '20210104T102623Z',
             'Content-Length': 0,
@@ -961,15 +962,15 @@ class TestRequest(S3ApiTestCase):
 
         # lowercase sha256
         req = Request.blank('/', environ=environ, headers=headers)
-        self.assertRaises(BadDigest, SigV4Request, req.environ)
+        self.assertRaises(XAmzContentSHA256Mismatch, SigV4Request, req.environ)
         sha256_of_nothing = hashlib.sha256().hexdigest().encode('ascii')
         headers = {
             'Authorization':
                 'AWS4-HMAC-SHA256 '
                 'Credential=test/20210104/us-east-1/s3/aws4_request, '
                 'SignedHeaders=host;x-amz-content-sha256;x-amz-date,'
-                'Signature=d90542e8b4c0d2f803162040a948e8e51db00b62a59ffb16682'
-                'ef433718fde12',
+                'Signature=96df261d8f0b617b7c6368e0c5d96ee61f1ec84005e826ece65'
+                'c0e0f97eba945',
             'X-Amz-Content-SHA256': sha256_of_nothing,
             'Date': 'Mon, 04 Jan 2021 10:26:23 -0000',
             'X-Amz-Date': '20210104T102623Z',
@@ -981,14 +982,14 @@ class TestRequest(S3ApiTestCase):
             sigv4_req._canonical_request().endswith(sha256_of_nothing))
         self.assertTrue(sigv4_req.check_signature('secret'))
 
-        # uppercase sha256
+        # uppercase sha256 -- signature changes, but content's valid
         headers = {
             'Authorization':
                 'AWS4-HMAC-SHA256 '
                 'Credential=test/20210104/us-east-1/s3/aws4_request, '
                 'SignedHeaders=host;x-amz-content-sha256;x-amz-date,'
-                'Signature=4aab5102e58e9e40f331417d322465c24cac68a7ce77260e9bf'
-                '5ce9a6200862b',
+                'Signature=7a3c396fd6043fb397888e6f4d6acc294a99636ff0bb57b283d'
+                '9e075ed87fce2',
             'X-Amz-Content-SHA256': sha256_of_nothing.upper(),
             'Date': 'Mon, 04 Jan 2021 10:26:23 -0000',
             'X-Amz-Date': '20210104T102623Z',
@@ -1000,12 +1001,194 @@ class TestRequest(S3ApiTestCase):
             sigv4_req._canonical_request().endswith(sha256_of_nothing.upper()))
         self.assertTrue(sigv4_req.check_signature('secret'))
 
+    @patch.object(S3Request, '_validate_dates', lambda *a: None)
+    def test_v4_req_xmz_content_sha256_mismatch(self):
+        # Virtual hosted-style
+        def fake_app(environ, start_response):
+            environ['wsgi.input'].read()
+
+        self.s3api.conf.storage_domains = ['s3.test.com']
+        environ = {
+            'HTTP_HOST': 'bucket.s3.test.com',
+            'REQUEST_METHOD': 'PUT'}
+        sha256_of_body = hashlib.sha256(b'body').hexdigest()
+        headers = {
+            'Authorization':
+                'AWS4-HMAC-SHA256 '
+                'Credential=test/20210104/us-east-1/s3/aws4_request, '
+                'SignedHeaders=host;x-amz-date,'
+                'Signature=5f31c77dbc63e7c6ffc84dae60a9261c57c44884fe7927baeb9'
+                '84f418d4d511a',
+            'Date': 'Mon, 04 Jan 2021 10:26:23 -0000',
+            'X-Amz-Date': '20210104T102623Z',
+            'Content-Length': 4,
+            'X-Amz-Content-SHA256': sha256_of_body,
+        }
+        req = Request.blank('/', environ=environ, headers=headers,
+                            body=b'not_body')
+        with self.assertRaises(XAmzContentSHA256Mismatch) as caught:
+            SigV4Request(req.environ).get_response(fake_app)
+        self.assertIn(b'<Code>XAmzContentSHA256Mismatch</Code>',
+                      caught.exception.body)
+        self.assertIn(
+            ('<ClientComputedContentSHA256>%s</ClientComputedContentSHA256>'
+             % sha256_of_body).encode('ascii'),
+            caught.exception.body)
+        self.assertIn(
+            ('<S3ComputedContentSHA256>%s</S3ComputedContentSHA256>'
+             % hashlib.sha256(b'not_body').hexdigest()).encode('ascii'),
+            caught.exception.body)
+
+    @patch.object(S3Request, '_validate_dates', lambda *a: None)
+    def test_v4_req_xmz_content_sha256_missing(self):
+        # Virtual hosted-style
+        self.s3api.conf.storage_domains = ['s3.test.com']
+        environ = {
+            'HTTP_HOST': 'bucket.s3.test.com',
+            'REQUEST_METHOD': 'PUT'}
+        headers = {
+            'Authorization':
+                'AWS4-HMAC-SHA256 '
+                'Credential=test/20210104/us-east-1/s3/aws4_request, '
+                'SignedHeaders=host;x-amz-date,'
+                'Signature=5f31c77dbc63e7c6ffc84dae60a9261c57c44884fe7927baeb9'
+                '84f418d4d511a',
+            'Date': 'Mon, 04 Jan 2021 10:26:23 -0000',
+            'X-Amz-Date': '20210104T102623Z',
+            'Content-Length': 0,
+        }
+        req = Request.blank('/', environ=environ, headers=headers)
+        self.assertRaises(InvalidRequest, SigV4Request, req.environ)
+
+    @patch.object(S3Request, '_validate_dates', lambda *a: None)
+    def test_v4_req_x_mz_content_sha256_bad_format(self):
+        # Virtual hosted-style
+        self.s3api.conf.storage_domains = ['s3.test.com']
+        environ = {
+            'HTTP_HOST': 'bucket.s3.test.com',
+            'REQUEST_METHOD': 'PUT'}
+        headers = {
+            'Authorization':
+                'AWS4-HMAC-SHA256 '
+                'Credential=test/20210104/us-east-1/s3/aws4_request, '
+                'SignedHeaders=host;x-amz-date,'
+                'Signature=5f31c77dbc63e7c6ffc84dae60a9261c57c44884fe7927baeb9'
+                '84f418d4d511a',
+            'Date': 'Mon, 04 Jan 2021 10:26:23 -0000',
+            'X-Amz-Date': '20210104T102623Z',
+            'Content-Length': 0,
+            'X-Amz-Content-SHA256': '0' * 63  # too short
+        }
+        req = Request.blank('/', environ=environ, headers=headers)
+        self.assertRaises(InvalidArgument, SigV4Request, req.environ)
+
+        headers['X-Amz-Content-SHA256'] = '0' * 63 + 'x'  # bad character
+        req = Request.blank('/', environ=environ, headers=headers)
+        self.assertRaises(InvalidArgument, SigV4Request, req.environ)
+
+    def test_validate_part_number(self):
+        sw_req = Request.blank('/nojunk',
+                               environ={'REQUEST_METHOD': 'GET'},
+                               headers={
+                                   'Authorization': 'AWS test:tester:hmac',
+                                   'Date': self.get_date_header()})
+        req = S3Request(sw_req.environ)
+        self.assertIsNone(req.validate_part_number())
+
+        # ok
+        sw_req = Request.blank('/nojunk?partNumber=102',
+                               environ={'REQUEST_METHOD': 'GET'},
+                               headers={
+                                   'Authorization': 'AWS test:tester:hmac',
+                                   'Date': self.get_date_header()})
+        req = S3Request(sw_req.environ)
+        self.assertEqual(102, req.validate_part_number())
+        req = S3Request(sw_req.environ,
+                        conf=Config({'max_upload_part_num': 100}))
+        self.assertEqual(102, req.validate_part_number(102))
+        req = S3Request(sw_req.environ,
+                        conf=Config({'max_upload_part_num': 102}))
+        self.assertEqual(102, req.validate_part_number(102))
+
+    def test_validate_part_number_invalid_argument(self):
+        def check_invalid_argument(part_num, max_parts, parts_count, exp_max):
+            sw_req = Request.blank('/nojunk?partNumber=%s' % part_num,
+                                   environ={'REQUEST_METHOD': 'GET'},
+                                   headers={
+                                       'Authorization': 'AWS test:tester:hmac',
+                                       'Date': self.get_date_header()})
+            req = S3Request(sw_req.environ,
+                            conf=Config({'max_upload_part_num': max_parts}))
+            with self.assertRaises(InvalidPartArgument) as cm:
+                req.validate_part_number(parts_count=parts_count)
+            self.assertEqual('400 Bad Request', str(cm.exception))
+            self.assertIn(
+                b'Part number must be an integer between 1 and %d' % exp_max,
+                cm.exception.body)
+
+        check_invalid_argument(102, 99, None, 99)
+        check_invalid_argument(102, 100, 99, 100)
+        check_invalid_argument(102, 100, 101, 101)
+        check_invalid_argument(102, 101, 100, 101)
+        check_invalid_argument(102, 101, 101, 101)
+        check_invalid_argument('banana', 1000, None, 1000)
+        check_invalid_argument(0, 10000, None, 10000)
+
+    def test_validate_part_number_invalid_part_number(self):
+        def check_invalid_part_num(part_num, max_parts, parts_count):
+            sw_req = Request.blank('/nojunk?partNumber=%s' % part_num,
+                                   environ={'REQUEST_METHOD': 'GET'},
+                                   headers={
+                                       'Authorization': 'AWS test:tester:hmac',
+                                       'Date': self.get_date_header()})
+            req = S3Request(sw_req.environ,
+                            conf=Config({'max_upload_part_num': max_parts}))
+            with self.assertRaises(InvalidPartNumber) as cm:
+                req.validate_part_number(parts_count=parts_count)
+            self.assertEqual('416 Requested Range Not Satisfiable',
+                             str(cm.exception))
+            self.assertIn(b'The requested partnumber is not satisfiable',
+                          cm.exception.body)
+
+        check_invalid_part_num(102, 10000, 1)
+        check_invalid_part_num(102, 102, 101)
+        check_invalid_part_num(102, 10000, 101)
+
+    def test_validate_part_number_with_range_header(self):
+        sw_req = Request.blank('/nojunk?partNumber=1',
+                               environ={'REQUEST_METHOD': 'GET'},
+                               headers={
+                                   'Range': 'bytes=1-2',
+                                   'Authorization': 'AWS test:tester:hmac',
+                                   'Date': self.get_date_header()})
+        req = S3Request(sw_req.environ)
+        with self.assertRaises(InvalidRequest) as cm:
+            req.validate_part_number()
+        self.assertEqual('400 Bad Request',
+                         str(cm.exception))
+        self.assertIn(b'Cannot specify both Range header and partNumber query '
+                      b'parameter', cm.exception.body)
+
+        # bad part number AND Range header
+        sw_req = Request.blank('/nojunk?partNumber=0',
+                               environ={'REQUEST_METHOD': 'GET'},
+                               headers={
+                                   'Range': 'bytes=1-2',
+                                   'Authorization': 'AWS test:tester:hmac',
+                                   'Date': self.get_date_header()})
+        req = S3Request(sw_req.environ)
+        with self.assertRaises(InvalidRequest) as cm:
+            req.validate_part_number()
+        self.assertEqual('400 Bad Request',
+                         str(cm.exception))
+        self.assertIn(b'Cannot specify both Range header and partNumber query '
+                      b'parameter', cm.exception.body)
+
 
 class TestSigV4Request(S3ApiTestCase):
     def setUp(self):
         super(TestSigV4Request, self).setUp()
         self.s3api.conf.s3_acl = True
-        self.swift.s3_acl = True
 
     def test_init_header_authorization(self):
         environ = {
@@ -1016,7 +1199,7 @@ class TestSigV4Request(S3ApiTestCase):
             x_amz_date = self.get_v4_amz_date_header()
             headers = {
                 'Authorization': auth,
-                'X-Amz-Content-SHA256': '0123456789',
+                'X-Amz-Content-SHA256': '0' * 64,
                 'Date': self.get_date_header(),
                 'X-Amz-Date': x_amz_date}
             req = Request.blank('/', environ=environ, headers=headers)
@@ -1047,7 +1230,7 @@ class TestSigV4Request(S3ApiTestCase):
             x_amz_date = self.get_v4_amz_date_header()
             headers = {
                 'Authorization': auth,
-                'X-Amz-Content-SHA256': '0123456789',
+                'X-Amz-Content-SHA256': '0' * 64,
                 'Date': self.get_date_header(),
                 'X-Amz-Date': x_amz_date}
             req = Request.blank('/', environ=environ, headers=headers)
@@ -1105,7 +1288,7 @@ class TestSigV4Request(S3ApiTestCase):
             x_amz_date = self.get_v4_amz_date_header()
             params['X-Amz-Date'] = x_amz_date
             signed_headers = {
-                'X-Amz-Content-SHA256': '0123456789',
+                'X-Amz-Content-SHA256': '0' * 64,
                 'Date': self.get_date_header(),
                 'X-Amz-Date': x_amz_date}
             req = Request.blank('/', environ=environ, headers=signed_headers,
@@ -1140,7 +1323,7 @@ class TestSigV4Request(S3ApiTestCase):
             x_amz_date = self.get_v4_amz_date_header()
             params['X-Amz-Date'] = x_amz_date
             signed_headers = {
-                'X-Amz-Content-SHA256': '0123456789',
+                'X-Amz-Content-SHA256': '0' * 64,
                 'Date': self.get_date_header(),
                 'X-Amz-Date': x_amz_date}
             req = Request.blank('/', environ=environ, headers=signed_headers,
@@ -1197,7 +1380,7 @@ class TestSigV4Request(S3ApiTestCase):
                 'Signature=X' % self.get_v4_amz_date_header().split('T', 1)[0])
         headers = {
             'Authorization': auth,
-            'X-Amz-Content-SHA256': '0123456789',
+            'X-Amz-Content-SHA256': '0' * 64,
             'Date': self.get_date_header(),
             'X-Amz-Date': x_amz_date}
 
@@ -1207,7 +1390,7 @@ class TestSigV4Request(S3ApiTestCase):
             return SigV4Request(req.environ, None, config)
 
         s3req = make_s3req(Config(), '/bkt', {'partNumber': '3'})
-        self.assertEqual(controllers.multi_upload.PartController,
+        self.assertEqual(controllers.ObjectController,
                          s3req.controller)
 
         s3req = make_s3req(Config(), '/bkt', {'uploadId': '4'})
@@ -1238,6 +1421,41 @@ class TestSigV4Request(S3ApiTestCase):
         self.assertEqual(controllers.ServiceController,
                          s3req.controller)
 
+    def test_controller_for_multipart_upload_requests(self):
+        environ = {
+            'HTTP_HOST': 'bucket.s3.test.com',
+            'REQUEST_METHOD': 'PUT'}
+        x_amz_date = self.get_v4_amz_date_header()
+        auth = ('AWS4-HMAC-SHA256 '
+                'Credential=test/%s/us-east-1/s3/aws4_request,'
+                'SignedHeaders=host;x-amz-content-sha256;x-amz-date,'
+                'Signature=X' % self.get_v4_amz_date_header().split('T', 1)[0])
+        headers = {
+            'Authorization': auth,
+            'X-Amz-Content-SHA256': '0' * 64,
+            'Date': self.get_date_header(),
+            'X-Amz-Date': x_amz_date}
+
+        def make_s3req(config, path, params):
+            req = Request.blank(path, environ=environ, headers=headers,
+                                params=params)
+            return SigV4Request(req.environ, None, config)
+
+        s3req = make_s3req(Config(), '/bkt', {'partNumber': '3',
+                                              'uploadId': '4'})
+        self.assertEqual(controllers.multi_upload.PartController,
+                         s3req.controller)
+
+        s3req = make_s3req(Config(), '/bkt', {'partNumber': '3'})
+        self.assertEqual(controllers.multi_upload.PartController,
+                         s3req.controller)
+
+        s3req = make_s3req(Config(), '/bkt', {'uploadId': '4',
+                                              'partNumber': '3',
+                                              'copySource': 'bkt2/obj2'})
+        self.assertEqual(controllers.multi_upload.PartController,
+                         s3req.controller)
+
 
 class TestHashingInput(S3ApiTestCase):
     def test_good(self):
@@ -1274,9 +1492,11 @@ class TestHashingInput(S3ApiTestCase):
         self.assertEqual(b'1234', wrapped.read(4))
         self.assertEqual(b'56', wrapped.read(2))
         # even though the hash matches, there was more data than we expected
-        with self.assertRaises(swob.HTTPException) as raised:
+        with self.assertRaises(S3InputSHA256Mismatch) as raised:
             wrapped.read(3)
-        self.assertEqual(raised.exception.status, '422 Unprocessable Entity')
+        self.assertIsInstance(raised.exception, BaseException)
+        # won't get caught by most things in a pipeline
+        self.assertNotIsInstance(raised.exception, Exception)
         # the error causes us to close the input
         self.assertTrue(wrapped._input.closed)
 
@@ -1288,9 +1508,8 @@ class TestHashingInput(S3ApiTestCase):
         self.assertEqual(b'1234', wrapped.read(4))
         self.assertEqual(b'56', wrapped.read(2))
         # even though the hash matches, there was more data than we expected
-        with self.assertRaises(swob.HTTPException) as raised:
+        with self.assertRaises(S3InputSHA256Mismatch):
             wrapped.read(4)
-        self.assertEqual(raised.exception.status, '422 Unprocessable Entity')
         self.assertTrue(wrapped._input.closed)
 
     def test_bad_hash(self):
@@ -1300,18 +1519,17 @@ class TestHashingInput(S3ApiTestCase):
             md5(raw, usedforsecurity=False).hexdigest())
         self.assertEqual(b'1234', wrapped.read(4))
         self.assertEqual(b'5678', wrapped.read(4))
-        with self.assertRaises(swob.HTTPException) as raised:
+        with self.assertRaises(S3InputSHA256Mismatch):
             wrapped.read(4)
-        self.assertEqual(raised.exception.status, '422 Unprocessable Entity')
         self.assertTrue(wrapped._input.closed)
 
     def test_empty_bad_hash(self):
-        wrapped = HashingInput(BytesIO(b''), 0, hashlib.sha256, 'nope')
-        with self.assertRaises(swob.HTTPException) as raised:
-            wrapped.read(3)
-        self.assertEqual(raised.exception.status, '422 Unprocessable Entity')
-        # the error causes us to close the input
-        self.assertTrue(wrapped._input.closed)
+        _input = BytesIO(b'')
+        self.assertFalse(_input.closed)
+        with self.assertRaises(XAmzContentSHA256Mismatch):
+            # Don't even get a chance to try to read it
+            HashingInput(_input, 0, hashlib.sha256, 'nope')
+        self.assertTrue(_input.closed)
 
 
 if __name__ == '__main__':
